#!/usr/pkg/bin/perl

# This is a reference "bare bones" implementation of botikkey client.
# For windowing OSes, a Tcl/Tk implementation with numerous bells and
# whistles is available.

$botikkey_port = "12040";

$INFO="/tmp/botik.info";

use IO::Socket;
BEGIN {push @INC, "/usr/lib/perl5/Botikkey";} use Digest::MD5 qw(md5_hex);

use vars qw ($verbose $password $sock $bkserver $mode $newmode $sock);

my $ipaddr_pat = '\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}';

sub usage ()
{
    die ("Usage: botikkey [-m WORLD] [-s server] [-v][-q] [password]\n".
	 "or     botikkey [-m WORLD] [-s server] [-v][-q] -f password-file\n".
         "MODE can be changed via SIG_USR1, SIG_USR2, SIGHUP\n");
}

sub parse_args ()
{
    while (defined ($_ = shift @ARGV)) {
	if (/^-v$/) {
	    $verbose++;
	}
	elsif (/^-q$/) {
	    $verbose = 0;
	}
	elsif (/^-f(.*)$/) {
	    usage () if defined $password;
	    my $file = $1 || shift @ARGV;
	    open (F, $file) || die "$file: $!\n";
	    $password = <F>;
            chomp $password;
	    close F;
	}
	elsif (/^-s(.*)$/) {
	    usage () if defined $bkserver;
            $bkserver = $1 || shift @ARGV || usage ();
	}
	elsif (/^-m(.*)$/) {
	    usage () if defined $mode;
            $mode = $1 || shift @ARGV || usage ();
            if ($mode !~ /^WORLD|LOCAL|NONE$/) {
                usage ();
            }
	}
	elsif (/^-/) {
	    usage ();
	}
	else {
	    usage () if defined $password;
	    $password = $_;
	}
    }
}

sub query_password ()
{
    select STDERR;
    $| = 1;
    system "stty -echo";
    print STDERR "Enter the password: ";
    $password = <>;
    chomp $password;
    print STDERR "\nOK\n";
    system "stty echo";
    select STDOUT;
    return $password;
}

sub on_sig_usr1
{
    $newmode = WORLD;
    warn "Got USR1, changed MODE to $newmode.\n" if $verbose;
}

sub on_sig_usr2
{
    $newmode = LOCAL;
    warn "Got USR2, changed MODE to $newmode.\n" if $verbose;
}

sub on_sig_hup
{
    $newmode = NONE;
    warn "Got HUP, changed MODE to $newmode.\n" if $verbose;
}
sub on_sig_int
{
    warn "Got SIGINT.\n" if $verbose;
    exit 0;
}
sub on_sig_term
{
    warn "Got SIGTERM.\n" if $verbose;
    #system "rm -f $pidfile";
    exit 0;
}

sub get_abon_info ()
{
    warn "Requesting abonent info...\n" if $verbose;
    $randstr   = rand().rand().rand();
    $random    = md5_hex($randstr);
    $dpassword = md5_hex($password);
    $digest    = md5_hex($random.$dpassword);
    $query     = "type=botikkey&random=".$random."&digest=".$digest;
    $url       = "https://informer.botik.ru/informer.fcgi?".$query;
    $res       = `wget '$url' -T 10 --no-check-certificate -q -O -`;
    $out       = "";
    for (split /\n/,$res) {
        if (/^org:(.*)$/) {
            $out .= "Abonent: " . $1 . "\n";
        }
        elsif (/^ip:(.*)$/) {
            $out .= "IP address: ". $1 . "\n";
        }
        elsif (/^name:(.*)$/) {
            $out .= "Hostname: " . $1 . "\n";
        }
        elsif (/^time:(.*)$/) {
            $out .= "Time: " . scalar localtime($1) . "\n";
        }
        elsif (/^left:(.*)$/) {
            $out .= "Balance: " . $1 . " rub.\n";
        }
        elsif (/^pb:(.*)$/) {
            $out .= "Blocking threshold: " . $1 . " rub.\n";
        }
        elsif (/^tariff:(.*)$/) {
            if ($1 eq "") {
                $t = "basic";
            } else {
                $t = $1;
            }
            $out .= "Tariff: " . $t . "\n";
        }
        elsif (/^nexttariff:(.*)$/) {
            if ($1 eq "") {
                $t = "basic";
            } else {
                $t = $1;
            }
            $out .= "Next tariff: " . $t . "\n";
        }
        elsif (/^tarifftime:(.*)$/) {
            if ($1 ne "") {
                $out .= "Tariff ends at: " . scalar localtime($1) . "\n";
            }
        }
        elsif (/^fromcache:(.*)$/) {
            # do not print this field
        }
        else {
            $out .= $_ . "\n";
        }
    }
    open F, ">", $INFO;
    print F $out;
    close F;
}

sub main ()
{
    parse_args ();
    $password = $ENV{BOTIKKEY} unless defined $password;
    #$password = query_password () unless defined $password;
    die "Please specify BKey password (or -f /pswd/file) in command line" unless defined $password;
    $dpassword = md5_hex ($password);
    $mode = WORLD unless $mode;
    system "echo $$ > /var/run/botikkey.pid";

    $SIG{PIPE} = 'IGNORE';
    $SIG{HUP} = \&on_sig_hup;
    $SIG{USR1} = \&on_sig_usr1;
    $SIG{USR2} = \&on_sig_usr2;
    $SIG{INT} = \&on_sig_int;
    $SIG{TERM} = \&on_sig_term;
    $itime = time(); # don't wget right after the daemon start

    # Don't die if server closes the socket, reconnect instead
reconnect:
    my $server = find_gateway ();
    if (! $server) {
        warn "No gateway found in routing table. Waiting...\n" if $verbose;
        sleep 30;
        goto reconnect;
    }

    $sock = IO::Socket::INET->new (Proto=>'tcp',
                                   PeerAddr=>$server,
                                   PeerPort=>$botikkey_port);
  if (!$sock) {
    warn "Couldn't connect to $server:$botikkey_port\n" if $verbose;
  } else {
	$sock->sockopt (SO_KEEPALIVE, 1);
	my $greeting = <$sock>;
	warn "Connected: $greeting" if $verbose;
	while (my $fromserver = <$sock>) {
        if ($newmode ne '') {
            $mode = $newmode;
            $newmode = '';
            close $sock;
            goto reconnect;
        }
	if ($fromserver =~ /^CHALLENGE\s+(\S+)/) {
			my $challenge = $1;
			print STDERR scalar localtime, " Got challenge... " if $verbose;

			my $salt = rand;
			my $digest = md5_hex ($challenge.$salt.$mode.$dpassword);
			warn "md5_hex ($challenge $salt $mode $dpassword)" if $verbose > 1;
			my $ok = print $sock "RESPONSE $digest $salt $mode\r\n";
			last if (! $ok);
			print STDERR "sent response... " if $verbose;

			my $reply = <$sock>;
			print STDERR "server replies: $reply" if $verbose;
	    }
	    elsif ($fromserver =~ /^MESSAGE\s+(.*)/) {
			my $message = $1;
			if (open F, "|wall") {
			    print F "Botikkey message received:\n";
			    for (split m% // %, $message) {
					print F;
			    }
			    close F || warn "wall exited with code $?\n";
			}
	    }
	    else {
			print $sock "Protocol violation\r\n";
			warn "Protocol violation, closing socket\n";
			last;
	    }
            if (time() - $itime >= 300) {
                # fetching abonent info every 5 minutes
                #get_abon_info ();
                $itime = time();
            }
	}
	warn "Lost connection.\n" if $verbose;
	close $sock;
  }

    sleep 3; # be polite
    goto reconnect;
}

sub find_gateway
{
    if (defined $bkserver) {
        return $bkserver;
    }
    open (F, "netstat -rn|") || die "netstat: $!";
    for (<F>) {
        if (/^(0\.0\.0\.0|default)\s+($ipaddr_pat)/) {
            close F;
            return $2;
        }
    }
    die "Cannot find default route in kernel routing table\n";
}

&main;
exit 0;

